#
# contrast.py -- mode for setting contrast
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np

from ginga.ImageView import ImageViewBase
from ginga.modes.mode_base import Mode


class ContrastMode(Mode):
    """
    Contrast Mode enables bindings that can adjust the contrast of
    an image in a Ginga image viewer.

    Enter the mode by
    -----------------
    * Space, then "t"

    Exit the mode by
    ----------------
    * Esc

    Default bindings in mode
    ------------------------
    * T : restore contrast to defaults
    * left drag : adjust contrast
      * Interactive shift/stretch colormap (aka contrast and bias).
      * Moving left/right controls shift, up/down controls stretch.
      * Release button when satisfied with the contrast.
    * right click : restore contrast to defaults
    * scroll wheel : change contrast (add Ctrl to change more finely)
    * Shift + scroll wheel : change brightness (add Ctrl to change more finely)
    * Ctrl + pan gesture : change contrast
    * Shift + pan gesture : change brightness

    """

    @classmethod
    def is_compatible_viewer(cls, viewer):
        return isinstance(viewer, ImageViewBase)

    def __init__(self, viewer, settings=None):
        super().__init__(viewer, settings=settings)

        self.actions = dict(
            dmod_contrast=['__t', None, None],

            kp_contrast_restore=['T', 'contrast+t', 'contrast+T'],

            sc_contrast=['contrast+scroll', 'contrast+*+scroll'],

            ms_contrast=['contrast+left', 'ctrl+right'],
            ms_contrast_restore=['contrast+right', 'ctrl+middle'],

            pa_contrast=['contrast+ctrl+pan', 'contrast+shift+pan'])

    def __str__(self):
        return 'contrast'

    @property
    def cancmap(self):
        bd = self.viewer.get_bindings()
        return bd.get_feature_allow('cmap')

    def start(self):
        pass

    def stop(self):
        self.onscreen_message(None)

    def restore_contrast(self, viewer, msg=True):
        viewer.get_settings().set(contrast=0.5, brightness=0.5)
        msg = self.settings.get('msg_cmap', msg)
        if msg:
            self.onscreen_message("Restored brightness and contrast",
                                  delay=0.5)
        return True

    def _tweak_colormap(self, viewer, x, y, mode):
        win_wd, win_ht = viewer.get_window_size()

        # translate Y cursor position as a percentage of the window
        # height into a contrast pct
        contrast_pct = np.clip(y, 0, win_ht) / float(win_ht)

        # translate X cursor position as a percentage of the window
        # width into a brightness pct
        brightness_pct = 1.0 - np.clip(x, 0, win_wd) / float(win_wd)

        with viewer.suppress_redraw:
            viewer.get_settings().set(contrast=contrast_pct,
                                      brightness=brightness_pct)

    def _change_contrast(self, viewer, msg, change_pct):
        msg = self.settings.get('msg_contrast', msg)
        pct = viewer.get_settings().get('contrast')
        change_factor = 1.0 + change_pct
        pct = min(1.0, max(0.0, pct * change_factor))
        viewer.get_settings().set(contrast=pct)
        if msg:
            self.onscreen_message("Contrast: %.2f%%" % (pct * 100),
                                  delay=1.0)

    def _change_brightness(self, viewer, msg, change_pct):
        msg = self.settings.get('msg_contrast', msg)
        pct = viewer.get_settings().get('brightness')
        change_factor = 1.0 + change_pct
        pct = min(1.0, max(0.0, pct * change_factor))
        viewer.get_settings().set(brightness=pct)
        if msg:
            self.onscreen_message("Brightness: %.2f%%" % (pct * 100),
                                  delay=1.0)

    #####  KEYBOARD ACTION CALLBACKS #####

    def kp_contrast_restore(self, viewer, event, data_x, data_y, msg=True):
        if not self.cancmap:
            return False
        event.accept()
        msg = self.settings.get('msg_cmap', msg)
        self.restore_contrast(viewer, msg=msg)

    #####  SCROLL ACTION CALLBACKS #####

    def sc_contrast(self, viewer, event, msg=True):
        """Interactively change the contrast or brightness by scrolling.
        """
        if not self.cancmap:
            return False
        event.accept()
        rev = self.settings.get('zoom_scroll_reverse', False)
        direction = self.get_direction(event.direction, rev=rev)
        change_pct = 0.05
        if 'ctrl' in event.modifiers:
            change_pct = 0.01
        if direction == 'down':
            change_pct = - change_pct
        if 'shift' in event.modifiers:
            self._change_brightness(viewer, msg, change_pct)
        else:
            self._change_contrast(viewer, msg, change_pct)

    #####  MOUSE ACTION CALLBACKS #####

    def ms_contrast(self, viewer, event, data_x, data_y, msg=True):
        """Shift the colormap by dragging the cursor left or right.
        Stretch the colormap by dragging the cursor up or down.
        """
        if not self.cancmap:
            return False
        event.accept()
        msg = self.settings.get('msg_contrast', msg)

        x, y = self.get_win_xy(viewer)

        if event.state == 'move':
            self._tweak_colormap(viewer, x, y, 'preview')

        elif event.state == 'down':
            self._start_x, self._start_y = x, y
            if msg:
                self.onscreen_message(
                    "Change contrast and brightness (drag mouse)", delay=1.0)
        else:
            self.onscreen_message(None)

    def ms_contrast_restore(self, viewer, event, data_x, data_y, msg=True):
        """An interactive way to restore the colormap contrast settings after
        a warp operation.
        """
        if not self.cancmap:
            return False
        event.accept()
        if event.state == 'down':
            self.restore_contrast(viewer, msg=msg)

    ##### GESTURE ACTION CALLBACKS #####

    def pa_contrast(self, viewer, event, msg=True):
        """Change the contrast or brightness by a pan gesture.
        (the back end must support gestures)
        """
        if not self.cancmap:
            return False
        event.accept()
        event = self._pa_synth_scroll_event(event)
        if event.state == 'move':
            rev = self.settings.get('zoom_scroll_reverse', False)
            direction = self.get_direction(event.direction, rev=rev)
            change_pct = 0.01
            if direction == 'down':
                change_pct = - change_pct
            if 'shift' in event.modifiers:
                self._change_brightness(viewer, msg, change_pct)
                return True
            elif 'ctrl' in event.modifiers:
                self._change_contrast(viewer, msg, change_pct)
                return True

        return False



#
# ap_region.py -- astropy-regions support
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""
This module provides Ginga support for DS9 type region files and objects via
the ``astropy-regions`` package.
"""
import numpy as np

from astropy import units as u
from astropy.coordinates import SkyCoord

from ginga.canvas.CanvasObject import get_canvas_types

HAVE_REGIONS = False
try:
    import regions
    HAVE_REGIONS = True
except ImportError:
    pass


__all__ = ['astropy_region_to_ginga_canvas_object', 'add_region',
           'ginga_canvas_object_to_astropy_region']


# mappings of point styles
pt_ginga = {'*': 'square', 'x': 'cross', '+': 'plus', 'D': 'diamond'}
pt_regions = {v: k for k, v in pt_ginga.items()}
# mappings of arrow styles
arr_ginga = {'0 0': 'none', '1 0': 'start', '0 1': 'end', '1 1': 'both'}
arr_regions = {v: k for k, v in arr_ginga.items()}


def astropy_region_to_ginga_canvas_object(r, logger=None):
    """
    Convert an astropy-region object to a Ginga canvas object.

    Parameters
    ----------
    r : subclass of `~regions.PixelRegion`
        The region object to be converted

    logger : a Python logger (optional, default: None)
        A logger to which errors will be written

    Returns
    -------
    obj : subclass of `~ginga.canvas.CanvasObject`
        The corresponding Ginga canvas object

    """
    if not HAVE_REGIONS:
        raise ValueError("Please install the Astropy 'regions' package to use this function")

    dc = get_canvas_types()
    obj = None
    if isinstance(r, (regions.CirclePixelRegion,)):
        obj = dc.Circle(r.center.x, r.center.y, r.radius)

    elif isinstance(r, (regions.CircleSkyRegion,)):
        obj = dc.Circle(r.center.ra.deg, r.center.dec.deg,
                        r.radius.to(u.deg).value, coord='wcs')

    elif isinstance(r, (regions.EllipsePixelRegion,)):
        obj = dc.Ellipse(r.center.x, r.center.y, r.width * 0.5, r.height * 0.5,
                         rot_deg=r.angle.to(u.deg).value)

    elif isinstance(r, (regions.EllipseSkyRegion,)):
        obj = dc.Ellipse(r.center.ra.deg, r.center.dec.deg,
                         (r.width * 0.5).to(u.deg).value,
                         (r.height * 0.5).to(u.deg).value,
                         rot_deg=r.angle.to(u.deg).value, coord='wcs')

    # NOTE: need to check for Text before Point, because Text seems to be
    # a subclass of Point in regions
    elif isinstance(r, (regions.TextPixelRegion,)):
        # NOTE: font needed here, but will be overridden later if specified
        # in the region's visuals
        obj = dc.Text(r.center.x, r.center.y, text=r.text, font='sans',
                      rot_deg=float(r.visual.get('textangle', 0.0)))

    elif isinstance(r, (regions.TextSkyRegion,)):
        # NOTE: font needed here, but will be overridden later if specified
        # in the region's visuals
        obj = dc.Text(r.center.ra.deg, r.center.dec.deg, text=r.text,
                      font='sans',
                      rot_deg=float(r.visual.get('textangle', 0.0)),
                      coord='wcs')

    elif isinstance(r, (regions.PointPixelRegion,)):
        # what is a reasonable default radius?
        radius = 15   # pixels

        # convert the regions-encoded style for a point into the
        # corresponding ginga style for a point, defaulting to "diamond"
        # if there is no direct match.
        style = r.visual.get('symbol', '*')
        style = pt_ginga.get(style, 'diamond')
        obj = dc.Point(r.center.x, r.center.y, radius, style=style)

    elif isinstance(r, (regions.PointSkyRegion,)):
        # what is a reasonable default radius?
        radius = 0.001   # degrees

        # see comment for PointPixelRegion
        style = r.visual.get('symbol', '*')
        style = pt_ginga.get(style, 'diamond')
        obj = dc.Point(r.center.ra.deg, r.center.dec.deg, radius, style=style,
                       coord='wcs')

    elif isinstance(r, (regions.LinePixelRegion,)):
        obj = dc.Line(r.start.x, r.start.y, r.end.x, r.end.y)
        obj.arrow = arr_ginga[r.meta.get('line', '0 0')]

    elif isinstance(r, (regions.LineSkyRegion,)):
        obj = dc.Line(r.start.ra.deg, r.start.dec.deg,
                      r.end.ra.deg, r.end.dec.deg, coord='wcs')
        obj.arrow = arr_ginga[r.meta.get('line', '0 0')]

    elif isinstance(r, (regions.RectanglePixelRegion,)):
        obj = dc.Box(r.center.x, r.center.y, r.width * 0.5, r.height * 0.5,
                     rot_deg=r.angle.to(u.deg).value)

    elif isinstance(r, (regions.RectangleSkyRegion,)):
        obj = dc.Box(r.center.ra.deg, r.center.dec.deg,
                     (r.width * 0.5).to(u.deg).value,
                     (r.height * 0.5).to(u.deg).value,
                     rot_deg=r.angle.to(u.deg).value, coord='wcs')

    elif isinstance(r, (regions.PolygonPixelRegion,)):
        points = np.array(r.vertices.xy).T
        obj = dc.Polygon(points)

    elif isinstance(r, (regions.PolygonSkyRegion,)):
        points = np.array((r.vertices.ra.deg, r.vertices.dec.deg)).T
        obj = dc.Polygon(points, coord='wcs')

    elif isinstance(r, (regions.CircleAnnulusPixelRegion,)):
        rin = r.inner_radius
        rout = r.outer_radius
        wd = rout - rin
        obj = dc.Annulus(r.center.x, r.center.y, rin, width=wd,
                         atype='circle')

    elif isinstance(r, (regions.CircleAnnulusSkyRegion,)):
        rin = r.inner_radius.to(u.deg).value
        rout = r.outer_radius.to(u.deg).value
        wd = rout - rin
        obj = dc.Annulus(r.center.ra.deg, r.center.dec.deg, rin, width=wd,
                         atype='circle', coord='wcs')

    elif isinstance(r, (regions.EllipseAnnulusPixelRegion,)):
        xwd = (r.outer_width - r.inner_width) * 0.5
        ywd = (r.outer_height - r.inner_height) * 0.5
        obj = dc.Annulus2R(r.center.x, r.center.y,
                           r.inner_width * 0.5, r.inner_height * 0.5,
                           xwidth=xwd, ywidth=ywd,
                           atype='ellipse',
                           rot_deg=r.angle.to(u.deg).value)

    elif isinstance(r, (regions.EllipseAnnulusSkyRegion,)):
        xwd = ((r.outer_width - r.inner_width) * 0.5).to(u.deg).value
        ywd = ((r.outer_height - r.inner_height) * 0.5).to(u.deg).value
        obj = dc.Annulus2R(r.center.ra.deg, r.center.dec.deg,
                           (r.inner_width * 0.5).to(u.deg).value,
                           (r.inner_height * 0.5).to(u.deg).value,
                           xwidth=xwd, ywidth=ywd,
                           atype='ellipse',
                           rot_deg=r.angle.to(u.deg).value, coord='wcs')

    elif isinstance(r, (regions.RectangleAnnulusPixelRegion,)):
        xwd = (r.outer_width - r.inner_width) * 0.5
        ywd = (r.outer_height - r.inner_height) * 0.5
        obj = dc.Annulus2R(r.center.x, r.center.y,
                           r.inner_width * 0.5, r.inner_height * 0.5,
                           xwidth=xwd, ywidth=ywd,
                           atype='box',
                           rot_deg=r.angle.to(u.deg).value)

    elif isinstance(r, (regions.RectangleAnnulusSkyRegion,)):
        xwd = ((r.outer_width - r.inner_width) * 0.5).to(u.deg).value
        ywd = ((r.outer_height - r.inner_height) * 0.5).to(u.deg).value
        obj = dc.Annulus2R(r.center.ra.deg, r.center.dec.deg,
                           (r.inner_width * 0.5).to(u.deg).value,
                           (r.inner_height * 0.5).to(u.deg).value,
                           xwidth=xwd, ywidth=ywd,
                           atype='box',
                           rot_deg=r.angle.to(u.deg).value, coord='wcs')

    else:
        errmsg = "Don't know how to convert this object of type: {}".format(str(type(r)))
        if logger is not None:
            # if a logger is passed, simply note the error message in the
            # log and convert the object to a Text with the error message
            logger.error(errmsg, exc_info=True)
            obj = dc.Text(0, 0, text=errmsg, font='sans')
        else:
            raise ValueError(errmsg)

    # Set visual styling attributes
    obj.color = r.visual.get('edgecolor', r.visual.get('color', 'green'))
    if hasattr(obj, 'font'):
        obj.font = r.visual.get('fontname', 'Sans')
        if 'fontsize' in r.visual:
            obj.fontsize = int(r.visual['fontsize'])

    if hasattr(obj, 'linewidth'):
        obj.linewidth = r.visual.get('linewidth', 1)

    if hasattr(obj, 'fill'):
        obj.fill = r.visual.get('fill', False)
        obj.fillcolor = r.visual.get('facecolor', obj.color)

    # Limited support for other metadata
    obj.editable = r.meta.get('edit', True)
    obj.set_data(name=r.meta.get('name', None))

    # needed for compound objects like annulus
    obj.sync_state()

    return obj


def add_region(canvas, r, tag=None, redraw=True):
    """
    Convenience function to plot an astropy-regions object on a Ginga
    canvas.

    Parameters
    ----------
    canvas : `~ginga.canvas.types.layer.DrawingCanvas`
        The Ginga canvas on which the region should be plotted.

    r : subclass of `~regions.PixelRegion`
        The region object to be plotted

    tag : str or None (optional, default: None)
        Caller can optionally pass a specific tag for the canvas object

    redraw : bool (optional, default: True)
        True if the viewers of the canvas should be updated immediately

    """
    obj = astropy_region_to_ginga_canvas_object(r)

    if tag is None:
        tag = obj.get_data('name')
    if obj is not None:
        canvas.add(obj, tag=tag, redraw=redraw)
        return obj


def ginga_canvas_object_to_astropy_region(obj, frame='icrs', logger=None):
    """
    Convert a Ginga canvas object to an astropy-region object.

    Parameters
    ----------
    obj : subclass of `~ginga.canvas.CanvasObject.CanvasObjectBase`
        The Ginga canvas object to be converted

    frame : str (optional, default: 'icrs')
        The type of astropy frame that should be generated for Sky regions

    logger : a Python logger (optional, default: None)
        A logger to which errors will be written

    Returns
    -------
    r : subclass of `~regions.PixelRegion` or `~regions.SkyRegion`
        The corresponding astropy-region object

    """
    if not HAVE_REGIONS:
        raise ValueError("Please install the Astropy 'regions' package to use this function")

    dc = get_canvas_types()
    r = None

    if isinstance(obj, (dc.Circle,)):
        if obj.coord in ('data', None):
            center = regions.PixCoord(x=obj.x, y=obj.y)
            r = regions.CirclePixelRegion(center=center,
                                          radius=obj.radius)
        elif obj.coord == 'wcs':
            center = SkyCoord(obj.x, obj.y, unit='deg', frame=frame)
            r = regions.CircleSkyRegion(center=center,
                                        radius=obj.radius * u.deg)

    elif isinstance(obj, (dc.Ellipse,)):
        if obj.coord in ('data', None):
            center = regions.PixCoord(x=obj.x, y=obj.y)
            r = regions.EllipsePixelRegion(center=center,
                                           width=obj.xradius * 2,
                                           height=obj.yradius * 2,
                                           angle=obj.rot_deg * u.deg)
        elif obj.coord == 'wcs':
            center = SkyCoord(obj.x, obj.y, unit='deg', frame=frame)
            r = regions.EllipseSkyRegion(center=center,
                                         width=obj.xradius * 2 * u.deg,
                                         height=obj.yradius * 2 * u.deg,
                                         angle=obj.rot_deg * u.deg)

    elif isinstance(obj, (dc.Text,)):
        if obj.coord in ('data', None):
            center = regions.PixCoord(x=obj.x, y=obj.y)
            r = regions.TextPixelRegion(center=center, text=obj.text)
            r.visual['textangle'] = str(obj.rot_deg)
        elif obj.coord == 'wcs':
            center = SkyCoord(obj.x, obj.y, unit='deg', frame=frame)
            r = regions.TextSkyRegion(center=center, text=obj.text)
            r.visual['textangle'] = str(obj.rot_deg)

    elif isinstance(obj, (dc.Point,)):
        style = pt_regions.get(obj.style, '*')
        if obj.coord in ('data', None):
            center = regions.PixCoord(x=obj.x, y=obj.y)
            r = regions.PointPixelRegion(center=center)
        elif obj.coord == 'wcs':
            center = SkyCoord(obj.x, obj.y, unit='deg', frame=frame)
            r = regions.PointSkyRegion(center=center)
        r.visual['symbol'] = style

    elif isinstance(obj, (dc.Line,)):
        if obj.coord in ('data', None):
            start = regions.PixCoord(x=obj.x1, y=obj.y1)
            end = regions.PixCoord(x=obj.x2, y=obj.y2)
            r = regions.LinePixelRegion(start=start, end=end)
        elif obj.coord == 'wcs':
            start = SkyCoord(obj.x1, obj.y1, unit='deg', frame=frame)
            end = SkyCoord(obj.x2, obj.y2, unit='deg', frame=frame)
            r = regions.LineSkyRegion(start=start, end=end)
        r.meta['line'] = arr_regions.get(obj.arrow, '0 0')

    elif isinstance(obj, (dc.Box,)):
        if obj.coord in ('data', None):
            center = regions.PixCoord(x=obj.x, y=obj.y)
            r = regions.RectanglePixelRegion(center=center,
                                             width=obj.xradius * 2,
                                             height=obj.yradius * 2,
                                             angle=obj.rot_deg * u.deg)
        elif obj.coord == 'wcs':
            center = SkyCoord(obj.x, obj.y, unit='deg', frame=frame)
            r = regions.RectangleSkyRegion(center=center,
                                           width=obj.xradius * 2 * u.deg,
                                           height=obj.yradius * 2 * u.deg,
                                           angle=obj.rot_deg * u.deg)

    elif isinstance(obj, (dc.Polygon,)):
        x, y = np.asarray(obj.points).T
        if obj.coord in ('data', None):
            vertices = regions.PixCoord(x=x, y=y)
            r = regions.PolygonPixelRegion(vertices=vertices)
        elif obj.coord == 'wcs':
            vertices = SkyCoord(x, y, unit='deg', frame=frame)
            r = regions.PolygonSkyRegion(vertices=vertices)

    elif isinstance(obj, (dc.Annulus,)) and obj.atype == 'circle':
        if obj.coord in ('data', None):
            rin = obj.radius
            rout = rin + obj.width
            center = regions.PixCoord(x=obj.x, y=obj.y)
            r = regions.CircleAnnulusPixelRegion(center=center,
                                                 inner_radius=rin,
                                                 outer_radius=rout)
        elif obj.coord == 'wcs':
            rin = obj.radius * u.deg
            rout = rin + obj.width * u.deg
            center = SkyCoord(obj.x, obj.y, unit='deg', frame=frame)
            r = regions.CircleAnnulusSkyRegion(center=center,
                                               inner_radius=rin,
                                               outer_radius=rout)

    elif isinstance(obj, (dc.Annulus2R,)) and obj.atype == 'ellipse':
        if obj.coord in ('data', None):
            center = regions.PixCoord(x=obj.x, y=obj.y)
            r = regions.EllipseAnnulusPixelRegion(center=center,
                                                  inner_width=obj.xradius * 2,
                                                  inner_height=obj.yradius * 2,
                                                  outer_width=obj.xradius * 2 + obj.xwidth * 2,
                                                  outer_height=obj.yradius * 2 + obj.ywidth * 2,
                                                  angle=obj.rot_deg * u.deg)
        elif obj.coord == 'wcs':
            center = SkyCoord(obj.x, obj.y, unit='deg', frame=frame)
            r = regions.EllipseAnnulusSkyRegion(center=center,
                                                inner_width=obj.xradius * 2 * u.deg,
                                                inner_height=obj.yradius * 2 * u.deg,
                                                outer_width=obj.xradius * 2 * u.deg + obj.xwidth * 2 * u.deg,
                                                outer_height=obj.yradius * 2 * u.deg + obj.ywidth * 2 * u.deg,
                                                angle=obj.rot_deg * u.deg)

    elif isinstance(obj, (dc.Annulus2R,)) and obj.atype == 'box':
        if obj.coord in ('data', None):
            center = regions.PixCoord(x=obj.x, y=obj.y)
            r = regions.RectangleAnnulusPixelRegion(center=center,
                                                    inner_width=obj.xradius * 2,
                                                    inner_height=obj.yradius * 2,
                                                    outer_width=obj.xradius * 2 + obj.xwidth * 2,
                                                    outer_height=obj.yradius * 2 + obj.ywidth * 2,
                                                    angle=obj.rot_deg * u.deg)
        elif obj.coord == 'wcs':
            center = SkyCoord(obj.x, obj.y, unit='deg', frame=frame)
            r = regions.RectangleAnnulusSkyRegion(center=center,
                                                  inner_width=obj.xradius * 2 * u.deg,
                                                  inner_height=obj.yradius * 2 * u.deg,
                                                  outer_width=obj.xradius * 2 * u.deg + obj.xwidth * 2 * u.deg,
                                                  outer_height=obj.yradius * 2 * u.deg + obj.ywidth * 2 * u.deg,
                                                  angle=obj.rot_deg * u.deg)

    else:
        errmsg = "Don't know how to convert this kind of object: {}".format(obj.kind)
        if logger is not None:
            # if a logger is passed, simply note the error message in the
            # log and convert the object to a TextPixelRegion with the
            # error message
            logger.error(errmsg, exc_info=True)
            r = regions.TextPixelRegion(center=regions.PixCoord(x=0, y=0),
                                        text=errmsg)
        else:
            raise ValueError(errmsg)

    # Set visual styling attributes
    r.visual['color'] = obj.color
    r.visual['edgecolor'] = obj.color

    if hasattr(obj, 'font'):
        r.visual['fontname'] = obj.font
        if obj.fontsize is not None:
            r.visual['fontsize'] = str(obj.fontsize)

    if hasattr(obj, 'linewidth'):
        r.visual['linewidth'] = obj.linewidth

    if hasattr(obj, 'fill'):
        r.visual['fill'] = 1 if obj.fill else 0
        r.visual['facecolor'] = obj.fillcolor

    # Limited support for other metadata
    r.meta['edit'] = 1 if obj.editable else 0
    meta = obj.get_data()
    if meta is not None and meta.get('name', None) is not None:
        r.meta['name'] = meta.get('name')

    return r


def import_regions(regions_file, format='ds9', logger=None):
    """
    Convenience function to read a file containing regions and
    return a list of matching Ginga canvas objects.

    Parameters
    ----------
    regions_file : str
        Path of a astropy-regions compatible file

    format : str (optional, default: 'ds9')
        Format of the astropy-regions compatible file

    logger : a Python logger (optional, default: None)
        A logger to which errors will be written

    Returns
    -------
    objs : list
        Returns a list of Ginga canvas objects that can be added
        to a Ginga canvas
    """
    regs = regions.Regions.read(regions_file, format=format)

    return [astropy_region_to_ginga_canvas_object(r, logger=logger)
            for r in regs]


def export_regions(objs, logger=None):
    """
    Convenience function to convert a sequence of Ginga canvas objects
    to a ds9 file containing regions and return a list of matching
    astropy-regions shapes.

    Parameters
    ----------
    objs : seq of subclasses of `~ginga.canvas.CanvasObject.CanvasObjectBase`
        Sequence of Ginga canvas objects compatible with Regions

    logger : a Python logger (optional, default: None)
        A logger to which errors will be written

    Returns
    -------
    regions : `~regions.Regions` object
        Returns an astropy-regions object
    """
    def _g2r(obj):
        return ginga_canvas_object_to_astropy_region(obj, logger=logger)
    regs = regions.Regions(map(_g2r, objs))
    return regs


def export_regions_canvas(canvas, logger=None):
    """
    Convenience function to convert a Ginga canvas's collection of objects
    to a ds9 file containing regions and return a list of matching
    astropy-regions shapes.

    Parameters
    ----------
    canvas : a `~ginga.canvas.types.layer.Canvas` object or subclass thereof
        a Ginga canvas object

    logger : a Python logger (optional, default: None)
        A logger to which errors will be written

    Returns
    -------
    regions : `~regions.Regions` object
        Returns an astropy-regions object
    """
    # TODO: support nested canvases, etc?
    def _g2r(obj):
        return ginga_canvas_object_to_astropy_region(obj, logger=logger)
    objs = canvas.objects
    return regions.Regions(map(_g2r, objs))



#
# GenericControl.py -- Generic controller base class.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
# stdlib imports
import sys
import os
import traceback
import tempfile
import threading
import webbrowser

# Local application imports
from ginga.misc import Bunch
from ginga.util import toolbox
from ginga.doc import download_doc

# GUI imports
from ginga.gw import GwHelp, GwMain, PluginManager
from ginga.gw import Widgets, Desktop
from ginga.util.paths import icondir as icon_dir

pluginconfpfx = None


class GenericShell(GwMain.GwMain, Widgets.Application):
    """
    Main generic shell for housing global plugins and building a UI from
    a layout.
    """
    def __init__(self, logger, thread_pool, module_manager, preferences,
                 ev_quit=None):
        GwMain.GwMain.__init__(self, logger=logger, ev_quit=ev_quit,
                               app=self, thread_pool=thread_pool)

        # Create general preferences
        self.prefs = preferences
        settings = self.prefs.create_category('general')
        settings.add_defaults(appname='ginga',
                              title="Ginga",
                              confirm_shutdown=True,
                              save_layout=False)
        settings.load(onError='silent')
        # this will set self.logger and self.settings
        Widgets.Application.__init__(self, logger=logger, settings=settings)

        self.mm = module_manager
        # event for controlling termination of threads executing in this
        # object
        if not ev_quit:
            self.ev_quit = threading.Event()
        else:
            self.ev_quit = ev_quit

        self.tmpdir = tempfile.gettempdir()

        # For callbacks
        for name in ['delete-workspace']:
            self.enable_callback(name)

        self.lock = threading.RLock()
        self.wscount = 0
        self.appname = settings.get('appname', 'ginga')

        self.plugins = []
        self._plugin_sort_method = self.get_plugin_menuname

        # This plugin manager handles "global" (aka standard) plug ins
        self.gpmon = self.get_plugin_manager(self.logger, self,
                                             None, self.mm)

        # GUI initialization
        self.w = Bunch.Bunch()
        self.main_wsname = None
        self._lastwsname = None
        self.ds = None
        self.layout = None
        self.layout_file = None
        self._help = Bunch.Bunch(remember_choice=False, choice=0)

        self.gui_dialog_lock = threading.RLock()
        self.gui_dialog_list = []
        self.w.root = None

    def get_preferences(self):
        return self.prefs

    def stop(self):
        self.logger.info(f"shutting down {self.appname}...")
        self.ev_quit.set()
        self.logger.debug("should be exiting now")

    # PLUGIN MANAGEMENT

    def get_plugin_manager(self, logger, image_viewer, ds, mm):
        return PluginManager.PluginManager(logger, image_viewer, ds, mm)

    def start_global_plugin(self, plugin_name, raise_tab=False):
        self.gpmon.start_plugin_future(None, plugin_name, None)
        if raise_tab:
            p_info = self.gpmon.get_plugin_info(plugin_name)
            self.ds.raise_tab(p_info.tabname)

    def stop_global_plugin(self, plugin_name):
        self.gpmon.deactivate(plugin_name)

    def call_global_plugin_method(self, plugin_name, method_name,
                                  args, kwargs):
        """
        Parameters
        ----------
        plugin_name : str
            The name of the global plugin containing the method to call.

        method_name : str
            The name of the method to call.

        args : list or tuple
            The positional arguments to the method

        kwargs : dict
            The keyword arguments to the method

        Returns
        -------
        result : return value from calling the method
        """
        p_obj = self.gpmon.get_plugin(plugin_name)
        method = getattr(p_obj, method_name)
        return self.gui_call(method, *args, **kwargs)

    def start_plugin(self, plugin_name, spec):
        self.start_global_plugin(plugin_name, raise_tab=True)

    def add_global_plugin(self, spec):
        try:
            spec.setdefault('ptype', 'global')
            name = spec.setdefault('name', spec.get('klass', spec.module))

            pfx = spec.get('pfx', pluginconfpfx)
            path = spec.get('path', None)
            self.plugins.append(spec)
            if spec.get('enabled', True):
                self.mm.load_module(spec.module, pfx=pfx, path=path)

            self.gpmon.load_plugin(name, spec)

        except Exception as e:
            self.logger.error("Unable to load global plugin '%s': %s" % (
                name, str(e)))

    def add_plugin(self, spec):
        self.add_global_plugin(spec)

    def set_plugins(self, plugins):
        self.plugins = []
        for spec in plugins:
            self.add_plugin(spec)

    def get_plugins(self):
        return self.plugins

    def get_plugin_spec(self, name):
        """Get the specification attributes for plugin with name `name`."""
        l_name = name.lower()
        for spec in self.plugins:
            name = spec.get('name', spec.get('klass', spec.module))
            if name.lower() == l_name:
                return spec
        raise KeyError(name)

    def get_plugin_menuname(self, spec):
        category = spec.get('category', None)
        name = spec.setdefault('name', spec.get('klass', spec.module))
        menu = spec.get('menu', spec.get('tab', name))
        if category is None:
            return menu
        return category + '.' + menu

    def set_plugin_sortmethod(self, fn):
        self._plugin_sort_method = fn

    def boot_plugins(self):
        # Sort plugins according to desired order
        self.plugins.sort(key=self._plugin_sort_method)

        for spec in self.plugins:
            name = spec.setdefault('name', spec.get('klass', spec.module))
            start = spec.get('start', True)
            # for now only start global plugins that have start==True
            if start and spec.get('ptype', 'local') == 'global':
                self.error_wrap(self.start_plugin, name, spec)

    def show_error(self, errmsg, raisetab=True):
        if self.gpmon.has_plugin('Errors'):
            obj = self.gpmon.get_plugin('Errors')
            obj.add_error(errmsg)
            if raisetab:
                self.ds.raise_tab('Errors')

    def error_wrap(self, method, *args, **kwargs):
        try:
            return method(*args, **kwargs)

        except Exception as e:
            errmsg = "\n".join([e.__class__.__name__, str(e)])
            try:
                (type, value, tb) = sys.exc_info()
                tb_str = "\n".join(traceback.format_tb(tb))
            except Exception as e:
                tb_str = "Traceback information unavailable."
            errmsg += tb_str
            self.logger.error(errmsg)
            self.gui_do(self.show_error, errmsg, raisetab=True)

    def help_text(self, name, text, text_kind='plain', trim_pfx=0):
        """
        Provide help text for the user.

        This method will trim the text as necessary and display it in
        the text widget.

        Parameters
        ----------
        name : str
            Category of help to show.

        text : str
            The text to show.

        text_kind : str (optional)
            One of 'plain' or 'rst'.  Default is 'plain'.

        trim_pfx : int (optional)
            Number of spaces to trim off the beginning of each line of text.

        """
        if trim_pfx > 0:
            # caller wants to trim some space off the front
            # of each line
            text = toolbox.trim_prefix(text, trim_pfx)

        if text_kind in ['rst', 'plain']:
            self.show_help_text(name, text)

        else:
            raise ValueError(
                "I don't know how to display text of kind '%s'" % (text_kind))

    def help(self, text=None, text_kind='url'):
        if text_kind == 'url':
            if text is None:
                # get top URL of external RTD docs
                text = download_doc.get_online_docs_url(plugin=None)
            self.show_help_url(text)
        else:
            if isinstance(text, str):
                self.show_help_text('HELP', text)

    def show_help_url(self, url):
        """
        Open a URL in an external browser using Python's webbrowser module.
        """
        self.logger.info(f"opening '{url}' in external browser...")
        webbrowser.open(url)

    def help_plugin(self, plugin_obj, text_kind='rst', url=None):
        """
        Called from a plugin's default help() method. Offers to show the
        user plugin docstring in a text widget or view the RTD doc in an
        external web browser.
        """

        def _do_help(val, url=None):
            if val == 1:
                # show plain text in a text widget
                if plugin_obj is not None:
                    name, doc = plugin_obj._get_docstring()
                    self.show_help_text(name, doc)
            elif val == 2:
                # show web page in external browser
                if url is None:
                    url = download_doc.get_online_docs_url(plugin=plugin_obj)
                self.show_help_url(url)

        if self._help.choice > 0:
            # User made a decision to keep getting plugin help the same way
            return _do_help(self._help.choice, url=url)

        # Create troubleshooting dialog if downloading cannot be done
        dialog = Widgets.Dialog(title="Show documentation",
                                parent=self.w.root,
                                modal=False,
                                buttons=[("Cancel", 0),
                                         ("Show RST text", 1),
                                         ("Use external browser", 2),
                                         ])
        dialog.buttons[0].set_tooltip("Skip help")
        dialog.buttons[1].set_tooltip("Show local docstring for plugin help")
        dialog.buttons[2].set_tooltip("Show online web documentation in external browser")
        vbox = dialog.get_content_area()
        dialog_text = Widgets.TextArea(wrap=True, editable=False)
        dialog_text.set_text("How would you like to see help?")
        vbox.add_widget(dialog_text, stretch=1)
        cb = Widgets.CheckBox("Remember my choice for session")
        cb.set_state(False)
        vbox.add_widget(cb, stretch=0)

        def _remember_choice_cb(w, tf):
            self._help.remember_choice = tf

        def _do_help_cb(dialog, val, url=None):
            if self._help.remember_choice:
                self._help.choice = val
            self.ds.remove_dialog(dialog)
            _do_help(val, url=url)

        cb.add_callback('activated', _remember_choice_cb)
        dialog.add_callback('activated', _do_help_cb, url=url)
        self.ds.show_dialog(dialog)

    def show_help_text(self, name, help_txt, wsname='channels'):
        """
        Show help text in a closeable tab window.  The title of the
        window is set from ``name`` prefixed with 'HELP:'
        """
        tabname = 'HELP: {}'.format(name)
        group = 1
        tabnames = self.ds.get_tabnames(group)
        if tabname in tabnames:
            # tab is already up somewhere
            return

        vbox = Widgets.VBox()
        vbox.set_margins(4, 4, 4, 4)
        vbox.set_spacing(2)

        msg_font = self.get_font('fixed', 12)
        tw = Widgets.TextArea(wrap=False, editable=False)
        tw.set_font(msg_font)
        tw.set_text(help_txt)
        vbox.add_widget(tw, stretch=1)

        btns = Widgets.HBox()
        btns.set_border_width(4)
        btns.set_spacing(3)

        def _close_cb(w):
            self.ds.remove_tab(tabname)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', _close_cb)
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)

        vbox.add_widget(btns, stretch=0)

        self.ds.add_tab(wsname, vbox, group, tabname)
        self.ds.raise_tab(tabname)

    def configure_workspace(self, wstype):
        ws = self.get_current_workspace()
        ws.configure_wstype(wstype)

    def cycle_workspace_type(self):
        ws = self.get_current_workspace()
        ws.cycle_wstype()

    def add_workspace(self, wsname, wstype, inSpace=None, use_toolbar=True):
        if inSpace is None:
            inSpace = self.main_wsname

        if wsname in self.ds.get_tabnames(None):
            raise ValueError("Tab name already in use: '%s'" % (wsname))

        ws = self.ds.make_ws(name=wsname, group=1, wstype=wstype,
                             use_toolbar=use_toolbar)
        if inSpace != 'top level':
            self.ds.add_tab(inSpace, ws.widget, 1, ws.name)
        else:
            #width, height = 700, 800
            #self.ds.create_toplevel_ws(width, height, group=1)
            top_w = self.ds.add_toplevel(ws, ws.name)
            ws.extdata.top_w = top_w

        return ws

    def delete_workspace(self, ws):
        # close the workspace
        top_w = ws.extdata.get('top_w', None)
        if top_w is None:
            self.ds.remove_tab(ws.name)
        else:
            # this is a top-level window
            self.ds.remove_toplevel(top_w)

        # inform desktop we are no longer tracking this
        self.ds.delete_ws(ws.name)

        self.make_gui_callback('delete-workspace', ws)

    def show_status(self, text):
        """Write a message to the status bar.

        Parameters
        ----------
        text : str
            The message.

        """
        self.status_msg("%s", text)

    def error(self, text):
        self.logger.error(text)
        self.status_msg("%s", text)
        # TODO: turn bar red

    def logit(self, text):
        try:
            obj = self.gpmon.get_plugin('Log')
            self.gui_do(obj.log, text)
        except Exception:
            pass

    def set_loglevel(self, level):
        handlers = self.logger.handlers
        for hdlr in handlers:
            hdlr.setLevel(level)

    def set_layout(self, layout, layout_file=None, save_layout=False,
                   main_wsname=None):
        self.layout = layout
        self.layout_file = layout_file
        self.save_layout = save_layout
        if main_wsname is not None:
            self.main_wsname = main_wsname

    def get_screen_dimensions(self):
        return (self.screen_wd, self.screen_ht)

    def build_toplevel(self, ignore_saved_layout=False):
        lo_file = self.layout_file
        if ignore_saved_layout:
            lo_file = None

        self.w.tooltips = None

        self.ds = Desktop.Desktop(self)
        self.ds.build_desktop(self.layout, lo_file=lo_file,
                              widget_dict=self.w)
        if self.main_wsname is None:
            ws = self.ds.get_default_ws()
            if ws is not None:
                self.main_wsname = ws.name
            else:
                # legacy value for layouts that don't define a default
                # workspace
                self.main_wsname = 'channels'
        self._lastwsname = self.main_wsname
        # TEMP: FIX ME!
        self.gpmon.ds = self.ds

        self.add_callback('close', self.close_cb)
        self.add_callback('shutdown', self.shutdown_cb)
        for win in self.ds.toplevels:
            # add delete/destroy callbacks
            win.add_callback('close', self.window_close)
            win.set_title(self.settings.get('title', "Ginga"))
            root = win
        self.ds.add_callback('all-closed', self.quit)

        self.w.root = root

        # initialize workspaces
        for wsname in self.ds.get_wsnames():
            ws = self.ds.get_ws(wsname)
            self.init_workspace(ws)

        if 'status' in self.w:
            statusholder = self.w['status']
            self.add_statusbar(statusholder)

        self.w.root.show()

    def add_statusbar(self, holder):
        self.w.status = Widgets.StatusBar()
        holder.add_widget(self.w.status, stretch=1)

    ####################################################
    # THESE METHODS ARE CALLED FROM OTHER MODULES & OBJECTS
    ####################################################

    def set_titlebar(self, text):
        self.w.root.set_title("{}: {}".format(self.appname.capitalize(), text))

    def status_msg(self, format, *args):
        if not format:
            s = ''
        else:
            s = format % args

        if 'status' in self.w:
            self.w.status.set_message(s)

    def set_pos(self, x, y):
        self.w.root.move(x, y)

    def set_size(self, wd, ht):
        self.w.root.resize(wd, ht)

    def set_geometry(self, geometry):
        # translation of X window geometry specification WxH+X+Y
        coords = geometry.replace('+', ' +')
        coords = coords.replace('-', ' -')
        coords = coords.split()
        if 'x' in coords[0]:
            # spec includes dimensions
            dim = coords[0]
            coords = coords[1:]
        else:
            # spec is position only
            dim = None

        if dim is not None:
            # user specified dimensions
            dim = list(map(int, dim.split('x')))
            self.set_size(*dim)

        if len(coords) > 0:
            # user specified position
            coords = list(map(int, coords))
            self.set_pos(*coords)

    def get_font(self, font_family, point_size):
        return GwHelp.get_font(font_family, point_size)

    def get_icon(self, icondir, filename):
        iconpath = os.path.join(icondir, filename)
        icon = GwHelp.get_icon(iconpath)
        return icon

    ####################################################
    # CALLBACKS
    ####################################################

    def window_close(self, w):
        """Quit the application.
        """
        # forces processing of close_cb
        self.close()

    def close_cb(self, app):
        if not self.settings.get('confirm_shutdown', True):
            self.quit()

        # confirm close with a dialog here
        q_quit = Widgets.Dialog(title="Confirm Quit", modal=False,
                                parent=self.w.root,
                                buttons=[("Cancel", False), ("Confirm", True)])
        # necessary so it doesn't get garbage collected right away
        self.w.quit_dialog = q_quit
        vbox = q_quit.get_content_area()
        vbox.set_margins(4, 4, 4, 4)
        hbox = Widgets.HBox()
        hbox.set_border_width(4)
        hbox.add_widget(Widgets.Label(""), stretch=1)
        img = Widgets.Image()
        iconfile = os.path.join(icon_dir, "warning.svg")
        img.load_file(iconfile)
        hbox.add_widget(img, stretch=0)
        hbox.add_widget(Widgets.Label(""), stretch=1)
        vbox.add_widget(hbox, stretch=1)
        vbox.add_widget(Widgets.Label("Do you really want to quit?"))
        q_quit.add_callback('activated', self._confirm_quit_cb)
        q_quit.add_callback('close', lambda w: self._confirm_quit_cb(w, False))
        q_quit.show()

    def _confirm_quit_cb(self, w, tf):
        dialog = self.w.quit_dialog
        self.w.quit_dialog = None
        if dialog is not None:
            dialog.delete()
        if not tf:
            return

        self.quit()

    def shutdown_cb(self, app):
        """Quit the application.
        """
        self.logger.info("Attempting to shut down the application...")

        # write out our current layout
        if self.layout_file is not None and self.save_layout:
            self.error_wrap(self.ds.write_layout_conf, self.layout_file)

        self.stop()

        # stop all global plugins
        self.gpmon.stop_all_plugins()

        self.w.root = None
        while len(self.ds.toplevels) > 0:
            w = self.ds.toplevels.pop()
            w.delete()

    def next_dialog(self):
        with self.gui_dialog_lock:
            # this should be the just completed call for a dialog
            # that gets popped off
            self.gui_dialog_list.pop(0)

            if len(self.gui_dialog_list) > 0:
                # if there are any other dialogs waiting, start
                # the next one
                future = self.gui_dialog_list[0]
                self.nongui_do_future(future)

    def init_workspace(self, ws):
        # subclass should override as necessary
        pass



#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""
The ``Pan`` plugin provides a small panning image that gives an overall
"birds-eye" view of the channel image that last had the focus.  If the
channel image is zoomed in 2X or greater, then the pan region is shown
graphically in the ``Pan`` image by a rectangle.

**Plugin Type: Local**

``Pan`` is a global plugin.  Only one instance can be opened.

**Usage**

The channel image can be panned by clicking and/or dragging to place
the rectangle.  Using the right mouse button to drag a rectangle will
force the channel image viewer to try to match the region (taking into
account the differences in the aspect ratio between the drawn rectangle
and the window dimensions).  Scrolling in the ``Pan`` image will zoom the
channel image.

The color/intensity map and cut levels of the ``Pan`` image are updated
when they are changed in the corresponding channel image.
The ``Pan`` image also displays the World Coordinate System (WCS) compass, if
valid WCS metadata is present in the FITS HDU being viewed in the
channel.

The ``Pan`` plugin usually appears as a sub-pane under the "Info" tab, next
to the ``Info`` plugin.

This plugin is not usually configured to be closeable, but the user can
make it so by setting the "closeable" setting to True in the configuration
file--then Close and Help buttons will be added to the bottom of the UI.

"""
import sys
import traceback
import math

from ginga.gw import Widgets, Viewers
from ginga.util import wcs
from ginga import GingaPlugin

__all__ = ['Pan']


class Pan(GingaPlugin.ParentPlugin):
    # Pan uses the ParentPlugin class to share the same window amongst
    # a bunch of local plugins

    def __init__(self, fv):
        super().__init__(fv)

        # the child plugin class we will start for each channel opened
        self.class_childplugin = PanChild

        spec = self.fv.get_plugin_spec(str(self))
        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_Pan')
        self.settings.add_defaults(closeable=not spec.get('hidden', False))
        self.settings.load(onError='silent')

        self.gui_up = False

    def build_gui(self, container):
        super().build_gui(container)

        if self.settings.get('closeable', False):
            btns = Widgets.HBox()
            btns.set_border_width(4)
            btns.set_spacing(4)

            btn = Widgets.Button("Close")
            btn.add_callback('activated', lambda w: self.close())
            btns.add_widget(btn)
            btn = Widgets.Button("Help")
            btn.add_callback('activated', lambda w: self.help())
            btns.add_widget(btn, stretch=0)
            btns.add_widget(Widgets.Label(''), stretch=1)

            container.add_widget(btns, stretch=0)
        self.gui_up = True

    def __str__(self):
        return 'pan'


class PanChild(GingaPlugin.ChildPlugin):
    # This local plugin actually does all the work, but only has to be
    # concerned with tracking a single channel

    def __init__(self, fv, image_viewer, parent_plugin):
        # superclass defines some variables for us, like logger
        super().__init__(fv, image_viewer, parent_plugin)

        self.dc = fv.get_draw_classes()

        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_Pan')
        self.settings.add_defaults(pan_position_color='yellow',
                                   pan_rectangle_color='red',
                                   compass_color='skyblue',
                                   rotate_pan_image=True)
        self.settings.load(onError='silent')

        self._wd = 200
        self._ht = 200
        self.copy_attrs = ['transforms', 'cutlevels', 'rotation', 'rgbmap',
                           'limits', 'icc', 'interpolation']

        self.compass_xy = None
        self.compass_wcs = None
        self.panrect = None

        self.gui_up = False

    def build_gui(self, container):
        vbox = Widgets.VBox()
        vbox.set_border_width(2)
        vbox.set_spacing(2)

        pi = Viewers.CanvasView(logger=self.logger)
        pi.enable_autozoom('on')
        pi.enable_autocuts('off')
        hand = pi.get_cursor('pan')
        pi.define_cursor('pick', hand)
        pi.set_bg(0.4, 0.4, 0.4)
        pi.set_desired_size(self._wd, self._ht)
        pi.set_callback('cursor-down', self.btndown)
        pi.set_callback('cursor-move', self.drag_cb)
        pi.set_callback('none-move', self.motion_cb)
        pi.set_callback('zoom-scroll', self.zoom_cb)
        pi.set_callback('zoom-pinch', self.zoom_pinch_cb)
        pi.set_callback('pan-pan', self.pan_pan_cb)
        pi.set_callback('configure', self.reconfigure)
        # for debugging
        pi.set_name('panimage')
        self.panimage = pi

        my_canvas = pi.get_private_canvas()
        my_canvas.enable_draw(True)
        my_canvas.set_drawtype('rectangle', linestyle='dash', color='green')
        my_canvas.set_callback('draw-event', self.draw_cb)

        # we share a canvas with the main channel image viewer
        canvas = self.fitsimage.get_canvas()
        pi.set_canvas(canvas)

        bd = pi.get_bindings()
        bd.enable_pan(False)
        bd.enable_zoom(False)

        p_canvas = pi.get_private_canvas()

        # add X/Y compass
        x, y = 0.5, 0.5
        radius = 0.1

        self.compass_xy = self.dc.Compass(x, y, radius,
                                          color=self.settings.get('xy_compass_color',
                                                                  'yellow'),
                                          fontsize=14, ctype='pixel',
                                          coord='percentage')
        p_canvas.add_object(self.compass_xy)

        self.compass_wcs = self.dc.Compass(x, y, radius * 2.0,
                                           color=self.settings.get('compass_color',
                                                                   'skyblue'),
                                           fontsize=14, ctype='wcs',
                                           coord='percentage')

        pw = Viewers.GingaViewerWidget(pi)
        pw.resize(self._wd, self._ht)

        self.fitsimage.copy_attributes(pi, self.copy_attrs)
        self.fitsimage.add_callback('redraw', self.redraw_cb)

        vbox.add_widget(pw, stretch=1)
        container.add_widget(vbox, stretch=1)
        self.gui_up = True

    def start(self):
        pass

    def stop(self):
        self.gui_up = False

    # CALLBACKS

    def redo(self):
        if not self.gui_up:
            return

        image = self.fitsimage.get_image()

        if (image is None) or not self.panimage.viewable(image):
            self.logger.debug("no main image--clearing Pan viewer")
            self.panimage.clear()
            return

        self.set_image(image)

    def blank(self, channel):
        if not self.gui_up:
            return
        self.panimage.clear()

    def reconfigure(self, panimage, width, height):
        self.logger.debug("new pan image dimensions are %dx%d" % (
            width, height))

        panimage.zoom_fit()
        panimage.redraw(whence=0)
        return True

    def redraw_cb(self, fitsimage, whence):
        if not self.gui_up:
            return
        if whence < 3:
            self.fitsimage.copy_attributes(self.panimage, self.copy_attrs,
                                           whence=whence)
            self.panimage.zoom_fit()
            self.panset()
        else:
            # nothing except a graphical overlay change, so simply
            # redraw to keep pace
            self.panimage.redraw(whence=whence)
        return True

    # LOGIC

    def clear(self):
        self.panimage.clear()

    def set_image(self, image):
        if image is None or not self.panimage.viewable(image):
            self.logger.debug("no main image--clearing Pan viewer")
            self.panimage.clear()
            return

        self.panimage.zoom_fit()

        p_canvas = self.panimage.get_private_canvas()
        # remove old WCS compass
        try:
            p_canvas.delete_object(self.compass_wcs)
        except Exception:
            pass

        # create compass
        if image.has_valid_wcs():
            try:
                x, y = 0.5, 0.5
                # HACK: force a wcs error here if one is going to happen
                wcs.add_offset_xy(image, x, y, 1.0, 1.0)

                p_canvas.add_object(self.compass_wcs)
            except Exception as e:
                self.logger.warning("Can't calculate wcs compass: {}".format(e))
                try:
                    # log traceback, if possible
                    (type_, value_, tb) = sys.exc_info()
                    tb_str = "".join(traceback.format_tb(tb))
                    self.logger.debug("Traceback:\n%s" % (tb_str))
                except Exception:
                    tb_str = "Traceback information unavailable."
                    self.logger.debug(tb_str)

        self.panset()

    def panset(self):
        image = self.fitsimage.get_image()
        if image is None or not self.panimage.viewable(image):
            self.panimage.clear()

        x, y = self.fitsimage.get_pan()
        points = self.fitsimage.get_pan_bbox()

        limits = self.fitsimage.get_limits()
        width = limits[1][0] - limits[0][0]
        height = limits[1][1] - limits[0][1]
        edgew = math.sqrt(width**2 + height**2)
        radius = int(0.015 * edgew)

        # Mark pan rectangle and pan position
        self.logger.debug("starting panset")
        p_canvas = self.panimage.get_private_canvas()
        if self.panrect is None:
            self.panrect = self.dc.CompoundObject(
                self.dc.Point(
                    x, y, radius, style='plus',
                    color=self.settings.get('pan_position_color', 'yellow')),
                self.dc.Polygon(
                    points,
                    color=self.settings.get('pan_rectangle_color', 'red')))
            p_canvas.add_object(self.panrect)
        else:
            point, bbox = self.panrect.objects
            point.x, point.y = x, y
            point.radius = radius
            bbox.points = points

        p_canvas.update_canvas(whence=3)

        #paninfo.panimage.zoom_fit()
        return True

    def get_channel_viewer(self):
        return self.channel.get_viewer('Ginga Image')

    def motion_cb(self, fitsimage, event, data_x, data_y):
        chviewer = self.get_channel_viewer()
        self.fv.showxy(chviewer, data_x, data_y)
        return True

    def drag_cb(self, fitsimage, event, data_x, data_y):
        # this is a panning move in the small
        # window for the big window
        chviewer = self.get_channel_viewer()
        chviewer.panset_xy(data_x, data_y)
        return True

    def btndown(self, fitsimage, event, data_x, data_y):
        chviewer = self.get_channel_viewer()
        chviewer.panset_xy(data_x, data_y)
        return True

    def zoom_cb(self, fitsimage, event):
        """Zoom event in the pan window.  Just zoom the channel viewer.
        """
        chviewer = self.get_channel_viewer()
        bd = chviewer.get_bindings()

        mode = bd.get_mode_obj('pan')
        return mode.sc_zoom(chviewer, event)

    def zoom_pinch_cb(self, fitsimage, event):
        """Pinch event in the pan window.  Just zoom the channel viewer.
        """
        chviewer = self.get_channel_viewer()
        bd = chviewer.get_bindings()

        mode = bd.get_mode_obj('pan')
        return mode.pi_zoom(chviewer, event)

    def pan_pan_cb(self, fitsimage, event):
        """Pan event in the pan window.  Just pan the channel viewer.
        """
        chviewer = self.get_channel_viewer()
        bd = chviewer.get_bindings()

        mode = bd.get_mode_obj('pan')
        return mode.pa_pan(chviewer, event)

    def draw_cb(self, canvas, tag):
        # Get and delete the drawn object
        obj = canvas.get_object_by_tag(tag)
        canvas.delete_object_by_tag(tag)

        # determine center of drawn rectangle and set pan position
        if obj.kind != 'rectangle':
            return False
        xc = (obj.x1 + obj.x2) / 2.0
        yc = (obj.y1 + obj.y2) / 2.0
        chviewer = self.get_channel_viewer()
        # note: chviewer <-- referring to channel viewer
        with chviewer.suppress_redraw:
            chviewer.panset_xy(xc, yc)

            # Determine appropriate zoom level to fit this rect
            wd = obj.x2 - obj.x1
            ht = obj.y2 - obj.y1
            wwidth, wheight = chviewer.get_window_size()
            wd_scale = float(wwidth) / float(wd)
            ht_scale = float(wheight) / float(ht)
            scale = min(wd_scale, ht_scale)
            self.logger.debug("wd_scale=%f ht_scale=%f scale=%f" % (
                wd_scale, ht_scale, scale))
            if scale < 1.0:
                zoomlevel = - max(2, int(math.ceil(1.0 / scale)))
            else:
                zoomlevel = max(1, int(math.floor(scale)))
            self.logger.debug("zoomlevel=%d" % (zoomlevel))

            chviewer.zoom_to(zoomlevel)

        return True

    def __str__(self):
        return 'panchild'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_Pan', package='ginga')

# END



#
# action.py -- Base classes for undo/redo actions in Ginga
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from collections import deque

#from ginga.misc import Bunch, Callback

__all__ = ['ActionStack', 'Action', 'AttrAction']


class ActionStack:

    def __init__(self):
        super().__init__()

        # TODO; add optional limit on stack size
        self._undo = deque()
        self._redo = deque()

    def push(self, action):
        self._undo.append(action)
        self._redo.clear()

    def pop(self):
        action = self._undo.pop()
        self._redo.append(action)
        return action

    def undo(self):
        action = self.pop()
        action.undo()
        return action

    def redo(self):
        action = self._redo.pop()
        self._undo.append(action)
        action.redo()
        return action


class Action:

    def __init__(self, descr=None):
        super(Action, self).__init__()

        self.description = descr

    def undo(self):
        pass

    def redo(self):
        pass


class AttrAction(Action):

    def __init__(self, obj, old, new, descr=None):
        """old and new are dicts mapping attribute names to values.
        """
        super().__init__(descr=descr)

        self.obj = obj
        self.old = old
        self.new = new

    def undo(self):
        for name, val in self.old.items():
            setattr(self.obj, name, val)

    def redo(self):
        for name, val in self.new.items():
            setattr(self.obj, name, val)



#
# evloop.py -- asyncio/tornado event loop utilities.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import sys
import asyncio


def get_event_loop():
    """Get the current asyncio event loop."""
    # workaround for DeprecationError in Python 3.10
    if sys.version_info < (3, 10):
        loop = asyncio.get_event_loop()
    else:
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

    return loop


def get_ioloop():
    """Get the current tornado ioloop."""
    from tornado.ioloop import IOLoop

    loop = get_event_loop()

    # now we should be able to get the current tornado IOLoop,
    # which as of Tornado > v5 is a wrapper around the asyncio
    # event loop
    return IOLoop.current()



"""Unit Tests for the colors.py functions"""

import logging

import pytest
from numpy.testing import assert_allclose

import ginga.colors


class TestColors(object):
    def setup_class(self):
        self.logger = logging.getLogger("TestColors")
        self.color_list_length = len(ginga.colors.color_dict)

    # Tests for the lookup_color() funtion

    @pytest.mark.parametrize(
        ('input_color', 'format', 'expected'),
        [("white", "tuple", (1.0, 1.0, 1.0)),
         ("black", "tuple", (0.0, 0.0, 0.0)),
         ("yellow", None, (1.0, 1.0, 0.0)),
         ("#ff0000", "tuple", (1, 0, 0)),
         ("#0000FF", "tuple", (0, 0, 1)),
         ("white", "hash", "#ffffff"),
         ("black", "hash", "#000000"),
         ("#FFFFFF", "hash", "#FFFFFF")])
    def test_lookup_color_to_tuple(self, input_color, format, expected):
        if format is not None:
            kwargs = {'format': format}
            if format == 'tuple':
                use_allclose = True
            else:
                use_allclose = False
        else:
            kwargs = {}
            use_allclose = True

        actual = ginga.colors.lookup_color(input_color, **kwargs)
        if use_allclose:
            assert_allclose(actual, expected)
        else:
            assert actual == expected

    @pytest.mark.parametrize(
        ('args', 'errtype'),
        [(("unknown_color", ), KeyError),
         (("unknown_key", ), KeyError),
         (("White", ), KeyError),
         (("#af8dc3", "name"), ValueError),
         (("white", "unknown_format"), ValueError)])
    def test_lookup_color_unknown(self, args, errtype):
        with pytest.raises(errtype):
            ginga.colors.lookup_color(*args)

    # Tests for the resolve_color() function

    @pytest.mark.parametrize(
        ('color_input', 'expected'),
        [('red', (1.0, 0.0, 0.0)),
         ('#FF0000', (1.0, 0.0, 0.0)),
         ('#00ff00', (0.0, 1.0, 0.0)),
         ((1.0, 0.0, 0.0), (1.0, 0.0, 0.0))])
    def test_resolve_color(self, color_input, expected):
        assert_allclose(ginga.colors.resolve_color(color_input), expected)

    # Tests for the get_colors() function

    def test_get_colors_len(self):
        expected = self.color_list_length
        actual = len(ginga.colors.get_colors())
        assert expected == actual

    def test_add_and_get_colors_len(self):
        ginga.colors.add_color("test_color_white", (0.0, 0.0, 0.0))

        expected = self.color_list_length + 1
        actual = len(ginga.colors.get_colors())
        assert expected == actual

        ginga.colors.remove_color("test_color_white")

    # Tests for the add_color() and remove_color() function

    def test_add_and_remove_color_len(self):
        ginga.colors.add_color("test_color_white", (0.0, 0.0, 0.0))

        expected = self.color_list_length + 1
        actual = len(ginga.colors.color_dict)
        assert expected == actual

        expected = len(ginga.colors.color_dict)
        actual = len(ginga.colors.color_list)
        assert expected == actual

        ginga.colors.remove_color("test_color_white")

        expected = self.color_list_length
        actual = len(ginga.colors.color_dict)
        assert expected == actual

        expected = len(ginga.colors.color_dict)
        actual = len(ginga.colors.color_list)
        assert expected == actual

    def test_add_and_remove_color_rbg(self):
        ginga.colors.add_color("test_color_white", (0.0, 0.0, 0.0))

        expected = (0.0, 0.0, 0.0)
        actual = ginga.colors.lookup_color("test_color_white")
        assert_allclose(expected, actual)

        ginga.colors.remove_color("test_color_white")
        with pytest.raises(KeyError):
            ginga.colors.remove_color("test_color_white")

    @pytest.mark.parametrize(
        ('colorname', 'colorval', 'errtype'),
        [("white", "string_wrong_format", TypeError),
         ("test_color", (-1.0, 0.0, 0.0), ValueError),
         ("test_color", (0.0, 0.0), ValueError)])
    def test_add_color_wrong(self, colorname, colorval, errtype):
        with pytest.raises(errtype):
            ginga.colors.add_color(colorname, colorval)

    def test_remove_color_unknown(self):
        with pytest.raises(KeyError):
            ginga.colors.remove_color("unknown_color")

    # Tests for recalc_color_list() function

    def test_recalc_color_list(self):
        ginga.colors.color_dict["test_color_white"] = (0.0, 0.0, 0.0)

        expected = len(ginga.colors.color_dict) - 1
        actual = len(ginga.colors.color_list)
        assert expected == actual

        ginga.colors.recalc_color_list()

        expected = len(ginga.colors.color_dict)
        actual = len(ginga.colors.color_list)
        assert expected == actual

        del ginga.colors.color_dict["test_color_white"]

        expected = len(ginga.colors.color_dict) + 1
        actual = len(ginga.colors.color_list)
        assert expected == actual

        ginga.colors.recalc_color_list()

        expected = len(ginga.colors.color_dict)
        actual = len(ginga.colors.color_list)
        assert expected == actual

    # Tests for scan_rgbtxt_buf() function

    def test_scan_rgbtxt_buf(self):
        test_rgb_lines = '''
                255 255 255		white
                0   0   0		black
                255   0   0		red
                0 255	  0		green
                0   0 255		blue
        '''

        result = ginga.colors.scan_rgbtxt_buf(test_rgb_lines)

        assert isinstance(result, dict)

        expected = 5
        actual = len(result)
        assert expected == actual

        expected = (1.0, 1.0, 1.0)
        actual = result["white"]
        assert_allclose(expected, actual)

# END
